/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: MultiPersistenceXMLManager.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.jpa;

import java.util.Iterator;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.ThreadSafe;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URL;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import org.apache.commons.io.IOUtils;
import it.tidalwave.util.logging.Logger;

/***********************************************************************************************************************
 *
 * This class is capable of creating a merged {@code persistence.xml} file obtained by properly putting together all the
 * {@code persistence.xml} files found in the system classloader.
 *
 * @author  Fabrizio Giudici
 * @version $Id: MultiPersistenceXMLManager.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
@ThreadSafe
public class MultiPersistenceXMLManager
  {
    private static final String CLASS = MultiPersistenceXMLManager.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private static final DocumentBuilderFactory DOC_BUILDER_FACTORY = DocumentBuilderFactory.newInstance();
        
    private static final XPathFactory XPATH_FACTORY = XPathFactory.newInstance();
        
    private static final XPath XPATH = XPATH_FACTORY.newXPath();
        
    private static final XPathExpression XPATH_ENTITY_PU_NODE;
//    private static final XPathExpression XPATH_ENTITY_PU_PROPERTIES_NODE;
//    private static final XPathExpression XPATH_PRIMARY_MARKER;
    private static final XPathExpression XPATH_ENTITY_CLASS_TEXT;
    
    private static final String PERSISTENCE_XML = "META-INF/persistence.xml";
    
    private static final String PRIMARY_MARKER = "it.tidalwave.netbeans.persistence.primaryPersistenceXml";

    private URL persistenceXmlUrl;
    
    static
      {
        try
          {
            XPATH_ENTITY_PU_NODE = XPATH.compile("/persistence/persistence-unit");
//            XPATH_ENTITY_PU_PROPERTIES_NODE = XPATH.compile("/persistence/persistence-unit/properties");
//            XPATH_PRIMARY_MARKER = XPATH.compile("/persistence/persistence-unit/properties/property" +
//                    "[@name='" + MASTER_MARKER +  "']/@value");
            XPATH_ENTITY_CLASS_TEXT = XPATH.compile("/persistence/persistence-unit/class/text()");
          }
        catch (XPathExpressionException e)
          {
            throw new ExceptionInInitializerError(e);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    enum Filter
      {
        PRIMARY
          {
            @Override
            public boolean filter (final @Nonnull Document document)
              throws XPathExpressionException
              {
                // FIXME: make the damn'd XPATH work
                Node node = document.getDocumentElement();
                try
                  {
                    final String toString = MultiPersistenceXMLManager.toString(node);
                    
//                    System.err.println("WWWWWWWWWWWWWW " + toString);
//                    Object x = XPATH_PRIMARY_MARKER.evaluate(document, XPathConstants.STRING);
//                    System.err.println("YYYYYYYYYYYY *" + x.getClass() + "*");

                    if (toString.contains(PRIMARY_MARKER))
                      {
//                          System.err.println("IS PRIMARY");
                        return true;
                      }
                  }
                catch (TransformerConfigurationException ex)
                  {
                    ex.printStackTrace();
                  }
                catch (TransformerException ex)
                  {
                    ex.printStackTrace();
                  }

                return false;
//                return XPATH_ENTITY_PU_PROPERTIES_NODE2.evaluate(document, XPathConstants.STRING) != null;
              }
          },
          
        OTHERS
          {
            @Override
            public boolean filter (final @Nonnull Document document)
              throws XPathExpressionException
              {
                return !PRIMARY.filter(document);
              }
          },

        ALL
          {
            @Override
            public boolean filter (final @Nonnull Document document)
              {
                return true;
              }
          };
        
        public abstract boolean filter (@Nonnull Document document)
          throws XPathExpressionException;
      }    
        
    /*******************************************************************************************************************
     *
     * Return the {@link URL} of a resource containing the aggregated {@code persistence.xml} file.
     *
     * @return   the URL of the aggregated {@code persistence.xml}
     *
     ******************************************************************************************************************/
    @Nonnull
    public synchronized URL getPersistenceXmlUrl()
      throws IOException 
      {
        if (persistenceXmlUrl == null)
          {
            PrintWriter pw = null;

            try 
              {
                final String persistenceXml = scanPersistenceXml();
                logger.fine("persistence.xml %s", persistenceXml);

                // WARNING: the base directory must be empty since Hibernate will scan it searching for classes,
                // so we are creating a brand new directory each time.
                final File folder = new File(System.getProperty("java.io.tmpdir") + "/PU-" + System.currentTimeMillis());
                folder.deleteOnExit();
                final File file = new File(folder, PERSISTENCE_XML);
                file.getParentFile().mkdirs();
                pw = new PrintWriter(new FileWriter(file));
                pw.print(persistenceXml);
                persistenceXmlUrl = file.toURL();
                logger.info("URL: %s", persistenceXmlUrl);
              } 
            catch (ParserConfigurationException e) 
              {
                throw new IOException(e.toString());
              } 
            catch (SAXException e) 
              {
                throw new IOException(e.toString());
              } 
            catch (XPathExpressionException e) 
              {
                throw new IOException(e.toString());
              } 
            catch (TransformerConfigurationException e)
              {
                throw new IOException(e.toString());
              } 
            catch (TransformerException e) 
              {
                throw new IOException(e.toString());
              }
            finally
              {
                IOUtils.closeQuietly(pw);
              }
          }

        return persistenceXmlUrl;
      }
    
    /*******************************************************************************************************************
     *
     * Scans for all the {@code persistence.xml} files in modules, merge them together and returns a {@link String}
     * representing the resulting XML document.
     *
     * @return  the merged XML document as a string
     *
     ******************************************************************************************************************/
    @Nonnull
    private String scanPersistenceXml()
      throws IOException, 
             ParserConfigurationException, 
             SAXException, 
             XPathExpressionException, 
             TransformerConfigurationException, 
             TransformerException
      {
        logger.fine("scanPersistenceXml()");
        final DocumentBuilder builder = DOC_BUILDER_FACTORY.newDocumentBuilder();
        DOC_BUILDER_FACTORY.setNamespaceAware(true); 
        final Iterator<URL> iterator = findPersistenceXmlFiles(Filter.PRIMARY).iterator();

        if (!iterator.hasNext())
          {
            final Collection<URL> persistenceXmlFiles = findPersistenceXmlFiles(Filter.ALL);
            throw new RuntimeException("Could not find a primary persistence.xml\n" +
                                       "A primary persistence.xml must contain a property named " + PRIMARY_MARKER + "\n" +
                                       "Found candidates: " + persistenceXmlFiles);
          }

        final URL masterURL = iterator.next();
        logger.fine(">>>> master persistence.xml: %s", masterURL);
        final Document masterDocument = builder.parse(masterURL.toExternalForm());
        final NodeList puNodes = (NodeList)XPATH_ENTITY_PU_NODE.evaluate(masterDocument, XPathConstants.NODESET);

        for (int n = 0; n < puNodes.getLength(); n++)
          {
            final Node puNode = puNodes.item(n);
            final Node firstChild = puNode.getFirstChild();
//            final Node propNode = (Node)XPATH_ENTITY_PU_PROPERTIES_NODE.evaluate(masterDocument, XPathConstants.NODE);
//
//            if (propNode != null)
//              {
//                propNode.getParentNode().removeChild(propNode);
//              }

            for (final URL url : findPersistenceXmlFiles(Filter.OTHERS))
              {
                logger.fine(">>>> secondary persistence.xml: %s", url);
                final Document document = builder.parse(url.toExternalForm());
                final NodeList nodes = (NodeList)XPATH_ENTITY_CLASS_TEXT.evaluate(document, XPathConstants.NODESET);

                for (int i = 0; i < nodes.getLength(); i++)
                  {
                    final String entityClassName = nodes.item(i).getNodeValue();
                    logger.fine(">>>>>>>> entity class: %s", entityClassName);

                    if (i == 0)
                      {
                        puNode.insertBefore(masterDocument.createTextNode("\n"), firstChild);
                        puNode.insertBefore(masterDocument.createComment(" from " +
                                url.toExternalForm().replaceAll(".*/cluster/modules/", "") + " "), firstChild);
                        puNode.insertBefore(masterDocument.createTextNode("\n"), firstChild);
                      }

                    final Node classNode = masterDocument.createElement("class");
                    classNode.appendChild(masterDocument.createTextNode(entityClassName));
                    puNode.insertBefore(classNode, firstChild);
                    puNode.insertBefore(masterDocument.createTextNode("\n"), firstChild);
                  }
              }

//            if (propNode != null)
//              {
//                puNode.appendChild(propNode);
//              }
          }
        
        return toString(masterDocument);
      }
    
    /*******************************************************************************************************************
     *
     * Finds the {@code persistence.xml} files reachable by the system classloader.
     *
     * @param  filter   whether we are searching the primary or the secondary instances
     * @return          a collection of URLs
     *
     ******************************************************************************************************************/
    @Nonnull
    private Collection<URL> findPersistenceXmlFiles (final @Nonnull Filter filter)
      throws IOException, ParserConfigurationException, SAXException, XPathExpressionException
      {
        final Collection<URL> result = new ArrayList<URL>();
        final ClassLoader classLoader = Thread.currentThread().getContextClassLoader().getParent();
        final DocumentBuilder builder = DOC_BUILDER_FACTORY.newDocumentBuilder();
        
        for (final Enumeration<URL> e = classLoader.getResources(PERSISTENCE_XML); e.hasMoreElements(); )
          {
            final URL url = e.nextElement();
            logger.finer(">>>> probing %s...", url);
            final Document document = builder.parse(url.toExternalForm());
            
            if (filter.filter(document))
              {
                result.add(url);
              }
          }
        
        return result;
      }
    
    /*******************************************************************************************************************
     *
     * Convenience method to convert a {@link Node} into a {@link String}.
     *
     * @param  node  the {@code Node}
     * @return       the string
     * 
     ******************************************************************************************************************/
    @Nonnull
    public static String toString (final @Nonnull Node node)
      throws TransformerConfigurationException, TransformerException 
      {
        final Source source = new DOMSource(node);
        final StringWriter stringWriter = new StringWriter();
        final Result result = new StreamResult(stringWriter);
        final TransformerFactory factory = TransformerFactory.newInstance();
        final Transformer transformer = factory.newTransformer();
        transformer.transform(source, result);
        return stringWriter.getBuffer().toString();
      }
  }
